"""
Sax content handlers used to parse scroll files (*.scrl).

MetaHandler is used by the Library to fill in the scroll's meta data attributes,
while the ContentHandler is used by the Scroll to fill in its contents.
"""

import xml.sax


class MetaHandler(xml.sax.ContentHandler):
    def __init__(self) -> None:
        self.meta_data = {}

    def get(self, key: any) -> any:
        return self.meta_data.get(key)

    def startElement(self, name, attrs):
        if name == "META":
            self.meta_data.update(attrs.items())


class ContentHandler(xml.sax.ContentHandler):
    def __init__(self) -> None:
        self.scroll_attributes = {}
        self._current_data = ""
        self.sections = {}
        self._section_content = []
        self.images = []
        self._paragraph_content = ""
        self.section_count = 0

    def startElement(self, tag, attributes):
        self._current_data = tag
        match self._current_data:
            case "section":
                self.section_count += 1
            case "img":
                self._section_content.append(f"IMGFILE:{attributes.get('src')}")
                self.images.append(f"{attributes.get('src')}")

    def endElement(self, tag):
        match tag:
            case "paragraph":
                self._paragraph_content += "\n\n"
                self._section_content.append(self._paragraph_content)
                self._paragraph_content = ""
            case "section":
                self.sections.update({self.section_count: self._section_content})
                self._section_content = []

    def characters(self, content):
        match self._current_data:
            case "paragraph":
                content = content.strip()
                if content.isprintable():
                    content_parts = content.rpartition("\n")
                    if content_parts[1] != "":
                        content = content_parts[0] + content_parts[2]
                    self._paragraph_content += content

    def endDocument(self):
        self.scroll_attributes.update({"sections": self.sections})



"""
Command line script to make a scroll file (*.scrl) from a well-formed directory.

Usage: scrollmake.py [-h] scroll_file_name target_dir
    scroll_file_name    Name of the scroll file.
    target_dir          Path of the well-formed directory.

    options:
    -h, --help          Show this help message and exit.

A well-formed directory should have a doctype scroll xml file following the pattern "*content*.xml"
along with any embedded media referenced in the contents file. The final *.scrl file will be placed
in the indicated library directory.
"""

import pathlib
import argparse
import tarfile
import shutil


def init_argparse() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog=__name__, description="Make a scroll file.")
    parser.add_argument("scroll_file_name", type=str, help="Name of scroll file.")
    parser.add_argument(
        "target_dir", type=str, help="Directory to turn into a scroll file."
    )

    return parser.parse_args()


def first_match(dir: pathlib.Path, pattern: str) -> pathlib.Path:
    return list(dir.glob(pattern))[0]


def main():
    args = init_argparse()

    current_dir = pathlib.Path(".")
    library_dir = first_match(current_dir, "scroll_library")
    target_dir = pathlib.Path(current_dir, args.target_dir)
    scroll_name = f"{args.scroll_file_name}.scrl"

    content = first_match(target_dir, "*content*.xml")
    media = list(target_dir.glob("*.jpg"))
    media.extend(list(target_dir.glob("*.png")))

    # TODO: Figure out how not to grab the entire path when adding a file to the archive.

    content = shutil.copyfile(content, pathlib.Path("./content.xml"))

    with tarfile.open(scroll_name, "x") as tar:
        tar.add(content)
        content.unlink()

        for file in media:
            file = shutil.copy(file, current_dir)
            file = pathlib.Path(file)
            tar.add(file)
            file.unlink()

    shutil.move(scroll_name, library_dir)


if __name__ == "__main__":
    main()



"""
Importable Scroller widgets.
"""

from typing import Any
import pytermgui as ptg
import pathlib
from ptgscroller.scroller import Library
from PIL import Image


class ScrollLibrary(ptg.Container):
    """
    A widget that generates a button for each scroll file in the library directory.

    Attributes:
        library (Library): Library instance to build the buttons from
        library_name (str): The name of the library instance
        window_manager (ptg.WindowManager): The PyTermGui windowmanager
    """

    def __init__(
        self,
        window_manager: ptg.WindowManager,
        library_dir: pathlib.Path = "./scroll_library",
        **attrs: Any,
    ) -> None:
        super().__init__(**attrs)
        attrs.update({"box": "EMPTY"})

        self.reader_embed_size = attrs.get("embed_size")
        if self.reader_embed_size is None:
            self.reader_embed_size = 32

        self.library = Library(library_dir)
        self.library_name = self.library.path.name
        self.window_manager = window_manager

        self.update_content()

    def _button_handler(self, scroll_title: str) -> callable:
        """
        Creates a function that creates a ScrollReader instance of the given scroll title.

        Args:
            scroll_title (str): The title of the scroll to open

        Returns:
            A function that opens a window with a ScrollReader of the given scroll
        """
        title = scroll_title

        def set_default_width() -> int:
            if self.reader_embed_size < 100:
                return 100
            else:
                return self.reader_embed_size + 50

        def draw_scroll_window(*_):
            reader = ScrollReader(self.library, title, self.reader_embed_size)
            exit = ptg.Button("X", lambda *_: window.close())

            window = (
                ptg.Window(exit, reader, width=set_default_width())
                .center()
                .set_title(reader.scroll_title)
            )

            self.window_manager.add(window)

        return draw_scroll_window

    def _generate_library_buttons(self) -> list[ptg.Button]:
        """
        Creates a list of PyTermGUI buttons for each scroll in a library directory that opens the corresponding scroll when clicked.

        Returns:
            A list of PyTermGUI buttons in PyTermGUI's auto syntax
        """
        library_buttons = []
        for title, scroll in self.library.scrolls.items():
            button = ptg.Button(title, self._button_handler(title))
            library_buttons.append(button)

        return library_buttons

    def update_content(self) -> None:
        """
        Create a container that contains generated library buttons and sets the ScrollLibrary's widgets.
        """
        buttons = ptg.Container(*self._generate_library_buttons(), box="EMPTY")

        self.set_widgets([buttons])


class ScrollReader(ptg.Container):
    """
    A widget that displays the content of a scroll file in a scrollable widget.

    Attributes:
        library (Library): Library instance to search through
        scroll_title (str): The title of the scroll to search for
        **attrs (Any): Attributes to pass to the widget
    """

    def __init__(
        self,
        library: Library,
        scroll_title: str,
        embed_size: int = 32,
        **attrs: Any,
    ) -> None:
        """
        Finds the corresponding scroll object using the scroll_title in the given library object, opens the scroll, and updates the widget's contents.
        """
        super().__init__(**attrs)
        attrs.update({"box": "EMPTY"})
        self.scroll_title = scroll_title
        self._scroll = library.find(scroll_title)

        self._scroll.embed_size = embed_size
        self._scroll.open()

        self.current_page = 1
        self.scroll_length = len(self._scroll.content.keys())

        self._inject_images()

        self._update_content()

    def _build_embed_image(self, image: Image):
        """
        Build a DensePixelMatrix from an Image object by setting matrix values according to the corresponding pixel value.

        Args:
            image (PIL.Image): An the Image object to turn into a DensePixelMatrix

        Returns:
            A built DensePixelMatrix
        """
        pixel_matrix = ptg.DensePixelMatrix(image.width, image.height, default="black")

        for horizontal in range(pixel_matrix.rows):
            for pixel in range(pixel_matrix.columns):
                pixel_value = image.getpixel((pixel, horizontal))

                if len(pixel_value) < 4:
                    r, g, b = pixel_value
                    pixel_matrix[horizontal, pixel] = f"{r};{g};{b}"
                else:
                    r, g, b, a = pixel_value

                    if a < 255:
                        pass
                    else:
                        pixel_matrix[horizontal, pixel] = f"{r};{g};{b}"

        pixel_matrix.build()

        return pixel_matrix

    def _inject_images(self):
        """
        Iterates through a scroll's content to find embedded image. Replaces it with the proper pixel matrix widget.
        """
        for page, scroll_content in self._scroll.content.items():
            for content in scroll_content:
                index = scroll_content.index(content)
                content_split = content.split(":")
                if content_split[0] == "IMGFILE":
                    image = self._scroll.images.get(content_split[1])
                    image_widget = self._build_embed_image(image)
                    content = image_widget

                scroll_content[index] = content
            self._scroll.content.update({page: scroll_content})

    def _get_section_content(self) -> list[str]:
        """
        Returns current page's corresponding content list from scroll's content dictionary.
        """

        return self._scroll.content.get(self.current_page)

    def _turn_page(self, page: int) -> None:
        """
        Turns the currently opened scroll to the given page.

        Args:
            page (int): Page to turn to
        """
        if page <= self.scroll_length and page > 0:
            self.current_page = page

        self._update_content()

    def _update_content(self) -> None:
        """
        Sets the ScrollReader's content viewer and page turning widgets.
        """
        content_viewer = ptg.Container(
            *self._get_section_content(),
            overflow=ptg.Overflow.SCROLL,
            parent_align=ptg.HorizontalAlignment.CENTER,
            height=20,
            box="EMPTY",
        )

        page_turner = ptg.Splitter(
            ["<", lambda *_: self._turn_page(self.current_page - 1)],
            f"{self.current_page}/{self.scroll_length}",
            [">", lambda *_: self._turn_page(self.current_page + 1)],
        )

        self.set_widgets([content_viewer, page_turner])



"""
Classes used by the Scroller package.
"""

import pathlib
import tempfile
import tarfile
import xml.sax
from PIL import Image
from ptgscroller.xml_handler import MetaHandler, ContentHandler


class Scroll:
    """
    Represents a scroll (*.scrl) file.

    The Scroll class provides functionality to read and navigate through the contents of a scroll file.

    Attributes:
        path (pathlib.Path): The path to the scroll file
        title (str): The title of the scroll
        meta (dict): A dictionary of meta attributes of a scroll ("Author", "Publish Date", etc)
        content (dict): The textual contents of the scroll file keyed by section number
    """

    def __init__(self, path: pathlib.Path) -> None:
        self.path = path
        self.title = ""
        self.meta = {}
        self.content = {}
        self.images = {}
        self.embed_size = 32

    def open(self):
        """
        Extracts the scroll file archive, parses the content.xml file, and inserts text content into content attribute.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            with tarfile.open(self.path) as tar:
                tar.extractall(tmpdir)

            tmpdir = pathlib.Path(tmpdir)
            content_file = list(tmpdir.glob("content.xml"))[0]

            handler = ContentHandler()
            parser = xml.sax.make_parser()
            parser.setContentHandler(handler)
            parser.parse(content_file)

            self.content = handler.scroll_attributes.get("sections")

            for image_name in handler.images:
                image_path = list(tmpdir.glob(image_name))[0]

                image = Image.open(image_path)
                image = image.resize((self.embed_size, self.embed_size))
                self.images.update({image_name: image})


class Library:
    """
    Represents a directory that contains a collection of scroll files.

    When instantiated scans the directory indicated in the path attribute for scroll files and populates scroll's metadata attributes.

    Attributes:
        scrolls (dict): Scroll title key and scroll object value
        path (pathlib.Path): Path to library directory
    """

    def __init__(self, path: pathlib.Path = "./scroll_library") -> None:
        self.scrolls = {}
        self.path = pathlib.Path(path)

        self.scan()

    def find(self, title: str) -> Scroll:
        """
        Get a scroll object from a library.

        Args:
            title (str): A scroll's title
        Returns:
            The corresponding scroll object matching the given title
        """
        return self.scrolls.get(title)

    def scan(self) -> None:
        """
        Look through library directory and populate for found scroll files. Fill in scroll file's metadata.
        """
        for scroll_tar in self.path.glob("*.scrl"):
            with tempfile.TemporaryDirectory() as tmpdir:
                with tarfile.open(scroll_tar) as tar:
                    tar.extractall(tmpdir)

                tmpdir = pathlib.Path(tmpdir)
                content_file = list(tmpdir.glob("content.xml"))[0]

                handler = MetaHandler()
                parser = xml.sax.make_parser()
                parser.setContentHandler(handler)
                parser.parse(content_file)

                scroll = Scroll(scroll_tar)
                scroll.title = handler.get("title")
                scroll.meta = handler.meta_data

                self.scrolls.update({scroll.title: scroll})
